/************************************************************

    CRITERIA.C

    08.10.2002

    Ismo Krkkinen

    Criteria for evaluating the goodness of clustering.

************************************************************/

#include "DistCrit/criteria.h"
#include "DistCrit/ftest.h"
#include "DistCrit/mdl.h"
#include "DistCrit/stcomplx.h"
#include "DistCrit/xiebenik.h"
#include "interfc.h"
#include "memctrl.h"
#include "random.h"
#include <math.h>
#include <string.h>

static void* InitMSE(DistanceInfo* DI, TRAININGSET* TS) {
    return (void*) allocate(sizeof(int) * BookSize(TS));
}

static void DoneMSE(void* AuxiliaryData) {
    deallocate(AuxiliaryData);
}

static float MeanError(DistanceInfo* DI, TRAININGSET* TS,
    CODEBOOK* CB, PARTITIONING* PA, int Squared)
{
    float Err = 0;
    int k;
    float Dist;
    for (k = 0; k < BookSize(TS); k++) {
        Dist = diDistance(DI, TS, CB, PA, k, Map(PA, k), DT_DATA2MODEL);
        if (DI->Type == DT_EuclideanSquare) {
            if (!Squared) Dist = (float)sqrt(Dist);
        } else if (Squared) Dist *= Dist;
        Err += Dist * VectorFreq(TS, k);
    }
    return Err / (TotalFreq(TS) * VectorSize(TS));
}

float CalculateMSE(DistanceInfo* DI, TRAININGSET* TS,
    CODEBOOK* CB, PARTITIONING* PA, void* UnusedParameter)
{
    return MeanError(DI, TS, CB, PA, 1);
}

float CalculateMAE(DistanceInfo* DI, TRAININGSET* TS,
    CODEBOOK* CB, PARTITIONING* PA, void* UnusedParameter)
{
    return MeanError(DI, TS, CB, PA, 0);
}

int MSECentroid(DistanceInfo* DI, TRAININGSET* TS,
    CODEBOOK* CB, PARTITIONING* PA, int* Changed, void* UnusedParameter)
{
    int k, n, Change, ChangeCount, NewScalar;
    llong ScalarSum;
    ChangeCount = 0;
    for (k = 0; k < BookSize(CB); k++) {
        ScalarSum = 0;
        Change = CCFreq(PA, k) != VectorFreq(CB, k);
        if (CCFreq(PA, k)) for (n = 0; n < VectorSize(CB); n++) {
            NewScalar = (int)
                (((double)CCScalar(PA, k, n) / (double)CCFreq(PA, k)) + 0.5);
            if (NewScalar != VectorScalar(CB, k, n)) {
                Change = 1;
                VectorScalar(CB, k, n) = NewScalar;
            }
            ScalarSum += VectorScalar(CB, k, n);
        } else { /* Empty cluster. */
            Change = irand(0, BookSize(TS) - 1);
            CopyVector(Vector(TS, Change), Vector(CB, k), VectorSize(TS));
            ChangePartition(TS, PA, k, Change);
            Change = 1;
        }
        if (Change) ++ChangeCount;
        Changed[k] = Change;
        VectorFreq(CB, k) = CCFreq(PA, k);
        Node(CB, k).vmean = (VECTORELEMENT)(ScalarSum / VectorSize(CB));
    }
    return ChangeCount;
}

int FindNearestCluster(DistanceInfo* DI, TRAININGSET* TS,
    CODEBOOK* CB, PARTITIONING* PA, int ForVector, int InCluster, int* Changed)
{
    int k, Index;
    float Dist, Nearest;
    Index = InCluster;
    Nearest = diDistance(DI, TS, CB, PA, ForVector, InCluster, DT_DATA2MODEL);
    for (k = 0; k < BookSize(CB); k++) if (Changed[InCluster] || Changed[k]) {
        Dist = diDistance(DI, TS, CB, PA, ForVector, k, DT_DATA2MODEL);
        if (Dist < Nearest) {
            Index = k;
            Nearest = Dist;
        }
    }
    return Index;
}

int MSEPartition(DistanceInfo* DI, TRAININGSET* TS,
    CODEBOOK* CB, PARTITIONING* PA, int* Changed, void* AuxiliaryData)
{
    int k, M, MoveCount;
    int* NewMap = (int*) AuxiliaryData;
    for (k = 0; k < BookSize(TS); k++)
        NewMap[k] = FindNearestCluster(DI, TS, CB, PA, k, Map(PA, k), Changed);
    MoveCount = 0;
    for (k = 0; k < BookSize(TS); k++) if (Map(PA, k) != NewMap[k]) {
        ++MoveCount;
        M = NewMap[k];
        Changed[M] = Changed[Map(PA, k)] = 1;
        VectorFreq(CB, Map(PA, k)) -= VectorFreq(TS, k);
        VectorFreq(CB, M) += VectorFreq(TS, k);
        ChangePartition(TS, PA, M, k);
    }
    return MoveCount;
}

int MSELocal(DistanceInfo* DI, TRAININGSET* TS,
    CODEBOOK* CB, PARTITIONING* PA, int* Changes, int Changed,
    void* UnusedParameter)
{
    int k, M, n, ChangeCount;
    float Dist, MinDist;
    ChangeCount = 0;
    for (k = 0; k < BookSize(TS); k++) {
        if (Map(PA, k) == Changed) {
            MinDist = diDistance(DI, TS, CB, PA, k, 0, DT_DATA2MODEL);
            M = 0;
            for (n = 1; n < BookSize(CB); n++) {
                Dist = diDistance(DI, TS, CB, PA, k, n, DT_DATA2MODEL);
                if (Dist < MinDist) {
                    MinDist = Dist;
                    M = n;
                }
            }
            if (M == Changed) continue;
            ChangeCount++;
            Changes[M] = Changes[Changed] = 1;
            VectorFreq(CB, Changed) -= VectorFreq(TS, k);
            VectorFreq(CB, M) += VectorFreq(TS, k);
            ChangePartition(TS, PA, M, k);
        } else if (
            diDistance(DI, TS, CB, PA, k, Map(PA, k), DT_DATA2MODEL) >
            diDistance(DI, TS, CB, PA, k, Changed, DT_DATA2MODEL))
        {
            ChangeCount++;
            Changes[Map(PA, k)] = Changes[Changed] = 1;
            VectorFreq(CB, Map(PA, k)) -= VectorFreq(TS, k);
            VectorFreq(CB, Changed) += VectorFreq(TS, k);
            ChangePartition(TS, PA, Changed, k);
        }
    }
    return ChangeCount;
}

CriterionInfo* ciNew(TRAININGSET* TS, DistanceInfo* DI,
    CR_TYPE Criterion, CR_InitializeFunction Init,
    CR_CriterionFunction Func, CR_FinalizeFunction Done,
    CR_PartitionFunction Partition, CR_LocalPartitionFunction Local,
    CR_CentroidFunction Centroid)
{
    CriterionInfo* ci = allocate(sizeof(CriterionInfo));
    memset(ci, 0, sizeof(CriterionInfo));
    ci->TS = TS;
    ci->Type = Criterion;
    ci->DI = DI;
    switch (Criterion) {
    case CR_User:
        if (!Func || !Partition || !Centroid || !Local) {
            ErrorMessage("criteria.c::ciNew: no user functions given.\n");
            ExitProcessing(FATAL_ERROR);
        }
        if ((!Init && Done) || (Init && !Done)) {
            ErrorMessage("criteria.c::ciNew: "
                "Both Init and Done must be given or neither.\n");
            ExitProcessing(FATAL_ERROR);
        }
        break;
    case CR_MSE:
        Init = &InitMSE;
        Done = &DoneMSE;
        Func = &CalculateMSE;
        Partition = &MSEPartition;
        Local = &MSELocal;
        Centroid = &MSECentroid;
        break;
    case CR_MAE:
        Init = &InitMSE;
        Done = &DoneMSE;
        Func = &CalculateMAE;
        Partition = &MSEPartition;
        Local = &MSELocal;
        Centroid = &MSECentroid;
        break;
    case CR_FTest:
        Init = &InitFTest;
        Done = &DoneFTest;
        Func = &CalculateFTest;
        Partition = &FTestPartition;
        Local = &MSELocal;
        Centroid = &MSECentroid;
        break;
    case CR_StochasticComplexity:
        Init = &InitStochasticComplexity;
        Done = &DoneStochasticComplexity;
        Func = &CalculateStochasticComplexity;
        Partition = &StochasticComplexityPartition;
        Local = &StochasticComplexityLocal;
        Centroid = &StochasticComplexityCentroid;
        break;
    case CR_MDL:
        Init = &InitMDL;
        Done = &DoneMDL;
        Func = &CalculateMDL;
        Partition = &MDLPartition;
        Local = &MDLLocal;
        Centroid = &MDLCentroid;
        break;
    case CR_XieBeniKwon:
        Init = &InitXieBeniKwon;
        Done = &DoneXieBeniKwon;
        Func = &CalculateXieBeniKwon;
        Partition = &XieBeniKwonPartition;
        Local = &MSELocal;
        Centroid = &MSECentroid;
        break;
    case CR_DBI:
        Init = &InitMSE;
        Done = &DoneMSE;
        Func = &CalculateDBI;
        Partition = MSEPartition;
        Local = &MSELocal;
        Centroid = &MSECentroid;
        break;
    default:
        ErrorMessage("criteria.c::ciNew: Unknown criterion.\n");
        ExitProcessing(FATAL_ERROR);
        break;
    }
    if (Init) ci->Data = Init(DI, TS);
    ci->Func = Func;
    ci->Done = Done;
    ci->Partition = Partition;
    ci->Local = Local;
    ci->Centroid = Centroid;
    return ci;
}

void ciDelete(CriterionInfo* CI) {
    if (CI->Done) (CI->Done)(CI->Data);
    deallocate(CI);
}

DT_TYPE ciCriterionDefaultDistance(CR_TYPE Criterion) {
    switch (Criterion) {
    case CR_User: return DT_User;
    case CR_MSE: return DT_EuclideanSquare;
    case CR_MAE: return DT_EuclideanSquare;
    case CR_FTest: return DT_EuclideanSquare;
    case CR_StochasticComplexity: return DT_DeltaSC;
    case CR_MDL: return DT_EuclideanSquare;
    case CR_XieBeniKwon: return DT_EuclideanSquare;
    case CR_DBI: return DT_EuclideanSquare;
    }
    ErrorMessage("criterion.c::CriterionDefaultDistance: bad criterion.\n");
    ExitProcessing(FATAL_ERROR);
    return 0;
}

