/***********************************************************************

    XIEBENIK.C

    11.10.2001

    Ismo Krkkinen

    Implements Kwon's variant of Xie-Beni -index.

***********************************************************************/

#include "DistCrit/xiebenik.h"
#include "memctrl.h"
#include "DistCrit/criteria.h"
#include <math.h>


static void MeanOfTS(TRAININGSET* TS, VECTORTYPE Result) {
    int n, k;
    llong* temp = allocate(sizeof(llong) * VectorSize(TS));
    memset(temp, 0, sizeof(llong) * VectorSize(TS));
    for (n = 0; n < BookSize(TS); n++) for (k = 0; k < VectorSize(TS); k++)
        temp[k] += VectorScalar(TS, n, k);
    for (k = 0; k < VectorSize(TS); k++) 
        Result[k] = (int)(temp[k] / BookSize(TS));
    deallocate(temp);
}

typedef struct {
    CODEBOOK Average;
    int* NewMap;
    CriterionInfo* CI;
} XieBeniKwonData;


void* InitXieBeniKwon(DistanceInfo* DI, TRAININGSET* TS) {
    XieBeniKwonData* D = allocate(sizeof(XieBeniKwonData));
    CreateNewCodebook(&(D->Average), 1, TS);
    MeanOfTS(TS, Vector(&(D->Average), 0));
    VectorFreq(&(D->Average), 0) = TotalFreq(TS);
    D->NewMap = allocate(sizeof(int) * BookSize(TS));
    D->CI = ciNew(TS, DI, CR_MSE, 0, 0, 0, 0, 0, 0);
    return (void*)D;
}

int XieBeniKwonPartition(DistanceInfo* DI, TRAININGSET* TS,
    CODEBOOK* CB, PARTITIONING* PA, int* Changed, void* AuxiliaryData)
{
    XieBeniKwonData* D = (XieBeniKwonData*)AuxiliaryData;
    return MSEPartition(DI, TS, CB, PA, Changed, D->NewMap);
}

float CalculateXieBeniKwon(DistanceInfo* DI, TRAININGSET* TS,
    CODEBOOK* CB, PARTITIONING* PA, void* AuxiliaryData)
{
    int k, n;
    float Nearest, Dist, CentroidMSE, SE;
    XieBeniKwonData* D = (XieBeniKwonData*)AuxiliaryData;
    SE = TotalFreq(TS) * ciEvaluate(D->CI, CB, PA);
    Nearest = diDistance(D->CI->DI, TS, CB, PA, 0, 1, DT_MODEL2MODEL);
    for (k = 0; k < BookSize(CB) - 1; ++k)
	for (n = k + 1; n < BookSize(CB); ++n) {
	    Dist = diDistance(D->CI->DI, TS, CB, PA, k, n, DT_MODEL2MODEL);
	    if (Dist < Nearest) Nearest = Dist;
	}
    CentroidMSE = 0;
    for (k = 0; k < BookSize(CB); k++) CentroidMSE +=
	diDistance(D->CI->DI, CB, &(D->Average), 0, k, 0, DT_DATA2MODEL);
    CentroidMSE /= BookSize(CB);
    if (Nearest > 0) return (VectorSize(CB) * SE + CentroidMSE) / Nearest;
    return 1e30;
}

void DoneXieBeniKwon(void* AuxiliaryData) {
    XieBeniKwonData* D = (XieBeniKwonData*)AuxiliaryData;
    FreeCodebook(&(D->Average));
    deallocate(D->NewMap);
    ciDelete(D->CI);
    deallocate(D);
}

/*==================== Davies-Bouldin Index ==========================*/

typedef struct {
    double Dispersion;
    double MaxW2BCS;
    int Contributor;
    int Moved;
    double Dist;
} DBIPartition;

static DBIPartition* GetDBIData(PARTITIONING* PA, CODEBOOK* CB,
    TRAININGSET* TS)
{
    DistanceInfo* DI = diNew(TS, DT_Euclidean, 0, 0, 0, 0);
    int k, n, Vect;
    DBIPartition* P;
    double W2BCS, Sum, Dist;
    P = allocate(sizeof(DBIPartition) * PartitionCount(PA));
    for (k = 0; k < PartitionCount(PA); k++) {

        P[k].Dispersion = 0;
        P[k].MaxW2BCS = 0;
        Sum = 0;

        Vect = FirstVector(PA, k);
        while (!EndOfPartition(Vect)) {
            Sum += diDistance(DI, TS, CB, PA, k, Vect, DT_MODEL2DATA);
            Vect = NextVector(PA, Vect);
        }

        P[k].Dispersion = Sum / (double) CCFreq(PA, k);

        for (n = 0; n < k; n++) {
            W2BCS = P[k].Dispersion + P[n].Dispersion;
            Dist = diDistance(DI, TS, CB, PA, k, n, DT_MODEL2MODEL);
            W2BCS /= Dist;
            if (W2BCS > P[k].MaxW2BCS) {
                P[k].MaxW2BCS = W2BCS;
                P[k].Contributor = n;
                P[k].Dist = Dist;
            }
            if (W2BCS > P[n].MaxW2BCS) {
                P[n].MaxW2BCS = W2BCS;
                P[n].Contributor = k;
                P[n].Dist = Dist;
            }
        }
    }
    diDelete(DI);
    return P;
}

static double DaviesBouldinIndex(PARTITIONING* PA, CODEBOOK* CB,
    TRAININGSET* TS)
{
    int k;
    DBIPartition* P;
    double Result = 0.0;
    P = GetDBIData(PA, CB, TS);
    for (k = 0; k < PartitionCount(PA); k++) Result += P[k].MaxW2BCS;
    deallocate(P);
    return Result / (double) PartitionCount(PA);
}


float CalculateDBI(DistanceInfo* DI, TRAININGSET* TS,
    CODEBOOK* CB, PARTITIONING* PA, void* AuxiliaryData)
{
    return DaviesBouldinIndex(PA, CB, TS);
}

