/***************************************************************

    CB_UTIL.C

    Ismo Krkkinen

    2.3.2001

    Miscellaneous functions for codebooks, training sets and
    partitions.

**************************************************************/

#include "cb_util.h"
#include "sortcb.h"
#include "memctrl.h"
#include "interfc.h"
#include "textfile.h"
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "random.h"
#include "file.h"
#include <limits.h>
#include "fvec.h"

void RemoveDuplicatesFromCodebook(CODEBOOK* CB) {
    int k, n;
    SortCodebook(CB, DATA_ASCENDING);
    k = 0;
    for (n = 1; n < BookSize(CB); n++) {
	if (EqualVectors(Vector(CB, k), Vector(CB, n), VectorSize(CB)))
	    VectorFreq(CB, k)++;
	else CopyVector(Vector(CB, n), Vector(CB, ++k), VectorSize(CB));
    }
    if (k + 1 < BookSize(CB)) DecreaseCodebookSize(CB, k + 1);
}

void GenerateRandomVector(TRAININGSET* TS, CODEBOOK* CB, int Index) {
    int x, n, Unique;
    do {
        Unique = 1;
        x = irand(0, BookSize(TS) - 1);
        for (n = 0; (n < BookSize(CB)) && Unique; n++) Unique = 
            !EqualVectors(Vector(TS, x), Vector(CB, n), VectorSize(CB));
    } while (!Unique);
    CopyVector(Vector(TS, x), Vector(CB, Index), VectorSize(CB));
    VectorFreq(CB, Index) = 0;
}

void GenerateRandomCodebook(TRAININGSET* TS, CODEBOOK* CB) {
    int k, n, x, Unique;
    for (k = 0; k < BookSize(CB); k++) {
	do {
	    Unique = 1;
	    x = irand(0, BookSize(TS) - 1);
	    for (n = 0; (n < k) && Unique; n++) Unique = 
		!EqualVectors(Vector(TS, x), Vector(CB, n), VectorSize(CB));
	} while (!Unique);
	CopyVector(Vector(TS, x), Vector(CB, k), VectorSize(CB));
	VectorFreq(CB, k) = 0;
    }
}


TRAININGSET* CreateTrainingSetFromRawData(char* Filename, 
                                          int FrequencyLocation,
                                          double Minimum, double Maximum,
                                          int Dimensionality, int Count,
                                          int BytesPerElement, int HasHeader,
                                          int QuietLevel)
{
    TRAININGSET* TS;
    int MaxVal, k, n, Error, Incomp, Dim;
    double* Vector, Range;
    FILE* F = fopen(Filename, "rt");
    TS = allocate(sizeof(TRAININGSET));
    if (BytesPerElement == 1) MaxVal = 255;
    else if (BytesPerElement == 2) MaxVal = 65535;
    else MaxVal = 1000000;
    CreateNewTrainingSet(TS, Count, Dimensionality, 1, BytesPerElement, 
	0, MaxVal, 1, "Translated from raw data.");
    Range = Maximum - Minimum;
    if (QuietLevel) PrintMessage("Range is [%f, %f].\n", Minimum, Maximum);
    k = 0;
    while (!feof(F) && k < Count) {
	Dim = FrequencyLocation ? Dimensionality + 1 : Dimensionality;
	Vector = ReadRawVector(F, &Dim, &Incomp, &Error);
	if (Vector && !Incomp && !Error) {
	    for (n = 0; n < Dim; n++) {
		if ((n == 0 && FrequencyLocation == 1)
		    || (n == Dim - 1 && FrequencyLocation == 2))
		{
		    VectorFreq(TS, k) = (int)Vector[n];
		} else if (FrequencyLocation == 1) 
		    VectorScalar(TS, k, n - 1) 
		        = (int)(((Vector[n] - Minimum) / Range) * MaxVal);
		else VectorScalar(TS, k, n) 
		    = (int)(((Vector[n] - Minimum) / Range) * MaxVal);
	    }
	    if (!FrequencyLocation) VectorFreq(TS, k) = 1;
	    TS->TotalFreq += VectorFreq(TS, k);
	    k++;
	}
        deallocate(Vector);
    }
    fclose(F);
    RemoveDuplicatesFromCodebook(TS);
    if (QuietLevel) 
	PrintMessage("Unique vectors after duplicate removal: %i\n\n",
	    BookSize(TS));
    return TS;
}

int CreateRawFile(char* OrigName, char* InName, char* OutName, CODEBOOK* CB, 
		  int FrequencyInfo, double Minimum, double Maximum,
		  int Header)
{
    double Val, Range;
    int k, OutputFrequencies, n;
    FILE* F = fopen(OutName, "wt");
    if (!F) return 0;
    if (Header) {
	/* output header info */
	fprintf(F, "%% Input file name: %s\n", InName);
	fprintf(F, "%% Raw data file name: %s\n", OrigName);
	fprintf(F, "%% Original range of values: %f - %f\n", Minimum, Maximum);
	fprintf(F, "%% Dimensionality: %i\n", VectorSize(CB));
	fprintf(F, "%% Generation method: %s\n", CB->GenerationMethod);
    }
    if (!FrequencyInfo && Header > 1) {
	OutputFrequencies = 0;
	for (k = 0; k < BookSize(CB); k++) if (VectorFreq(CB, k) > 2) {
	    OutputFrequencies = 1;
	    break;
	}
	if (OutputFrequencies) {
	    fprintf(F, "%% Frequencies:\n");
	    for (k = 0; k < BookSize(CB); k++)
		fprintf(F, "%% %i\n", VectorFreq(CB, k));
	}
    }
    /* output the data */
    Range = Maximum - Minimum;
    for (k = 0; k < BookSize(CB); k++) {
	if (FrequencyInfo == 1) fprintf(F, "%i ", VectorFreq(CB, k));
	for (n = 0; n < VectorSize(CB); n++) {
	    Val = (float)VectorScalar(CB, k, n) / (float)CB->MaxValue;
	    Val = (Val * Range) + Minimum;
	    fprintf(F, "%f ", Val);
	}
	if (FrequencyInfo == 2) fprintf(F, "%i ", VectorFreq(CB, k));
	fprintf(F, "\n");
    }
    fclose(F);
    return 1;
}

static int IsCodebookFile(char* Filename) {
    char Buffer[MaxVersionLength];
    FILE* f = FileOpen(Filename, INPUT, NO);
    fgets(Buffer, MaxVersionLength, f);
    fclose(f);
    while (isspace((int)Buffer[strlen(Buffer) - 1]))
	Buffer[strlen(Buffer) - 1] = 0;
    if (strcmp("VQ CODEBOOK 2.0", Buffer) &&
	strcmp("VQ TRAINING SET 2.0", Buffer)) return 0;
    return 1;
}

int ReadGenericData(char* Filename, TRAININGSET* TS, int* Format,
    float* Minimum, float* Maximum)
{
    FILE* f;
    int k, n, Count, Dim, OK;
    float** Data;
    float Scale;
    if (IsCodebookFile(Filename)) {
	ReadTrainingSet(Filename, TS);
	*Format = RGD_CB;
	*Minimum = TS->MinValue;
	*Maximum = TS->MaxValue;
	if (TS->MaxValue == 1 && TS->MinValue == 0) {
	    for (n = 0; n < BookSize(TS); n++)
		for (k = 0; k < VectorSize(TS); k++)
		    if (VectorScalar(TS, n, k))
			VectorScalar(TS, n, k) = 1000000;
	    TS->MaxValue = 1000000;
	    TS->BytesPerElement = 3;
	}
    } else {
	f = fopen(Filename, "rt");
	OK = fvReadSet(&Data, &Count, &Dim, f);
	fclose(f);
	if (!OK) return 1;
	*Format = RGD_TXT;
	*Minimum = fvSetTotalMinimum(Data, Count, Dim);
	*Maximum = fvSetTotalMaximum(Data, Count, Dim);
	if (fvSetIsBinary(Data, Count, Dim, *Minimum, *Maximum)) {
	    Scale = 1000000 / (*Maximum - *Minimum);
	    CreateNewTrainingSet(TS, Count, Dim, 1, 4, 0, Count, 0, "RGD");
	} else {
	    Scale = INT_MAX / (*Maximum - *Minimum);
	    CreateNewTrainingSet(TS, Count, Dim, 1, 4, 0, INT_MAX, 0, "RGD");
	}
	for (n = 0; n < Count; n++) {
	    VectorFreq(TS, n) = 1;
	    for (k = 0; k < Dim; k++)
		VectorScalar(TS, n, k) = (Data[n][k] - *Minimum) * Scale;
	}
	TotalFreq(TS) = Count;
	fvDeleteSet(Data, Count);
    }
    return 0;
}

static int IsPartitioningFile(char* Filename) {
    char Buffer[MaxVersionLength];
    FILE* f = FileOpen(Filename, INPUT, NO);
    fgets(Buffer, MaxVersionLength, f);
    fclose(f);
    while (isspace((int)Buffer[strlen(Buffer) - 1]))
	Buffer[strlen(Buffer) - 1] = 0;
    if (strcmp("VQ PARTITIONING 2.0", Buffer)) return 0;
    return 1;
}

int ReadGenericMapping(char* Filename, TRAININGSET* TS, PARTITIONING* PA,
    int* Format)
{
    FILE* f;
    int k, n, Count, Dim, Uniques, OK, M;
    float** Mapping;
    float* Unis;
    if (IsPartitioningFile(Filename)) {
	ReadPartitioning(Filename, PA, TS);
	*Format = RGD_CB;
    } else {
	f = fopen(Filename, "rt");
	OK = fvReadSet(&Mapping, &Count, &Dim, f);
	fclose(f);
	if (!OK) return 1;
	*Format = RGD_TXT;
	if ((Count != BookSize(TS)) || (Dim != 1)) {
	    fvDeleteSet(Mapping, Count);
	    ErrorMessage("Not suitable mapping data.\n");
	    return 1;
	}
	Unis = fvSetUnique(Mapping, Count, 0, &Uniques);
	for (k = 0; k < Uniques; k++) if (Unis[k] > Uniques - 1) {
	    ErrorMessage("Too high index to group.\n");
	    deallocate(Unis);
	    fvDeleteSet(Mapping, Count);
	    return 0;
	}
	deallocate(Unis);
	CreateNewPartitioning(PA, TS, Uniques);
        for (k = 0; k < PartitionCount(PA); k++) {
            FirstVector(PA, k) = ENDPARTITION;
            for (n = 0; n < VectorSize(TS); n++) CCScalar(PA, k, n) = 0;
            CCFreq(PA, k) = 0;
            UniqueVectors(PA, k) = 0;
        }
	for (k = 0; k < BookSize(TS); k++) {
	    Map(PA, k) = M = (int)Mapping[k][0];
	    NextVector(PA, k) = FirstVector(PA, M);
	    FirstVector(PA, M) = k;
	    for (n = 0; n < VectorSize(TS); n++) CCScalar(PA, M, n)
		+= VectorFreq(TS, k) * (llong) VectorScalar(TS, k, n);
	    CCFreq(PA, M) += VectorFreq(TS, k);
	    UniqueVectors(PA, M)++;
	}
	fvDeleteSet(Mapping, Count);
    }
    return 0;
}

void WriteGenericData(char* Filename, CODEBOOK* CB, int Format,
    float Minimum, float Maximum, int AllowOverwrite)
{
    int n, k;
    FILE* f;
    if (!AllowOverwrite && ExistFile(Filename)) {
        ErrorMessage("File\n%s\nexists and no overwrite allowed.\n", Filename);
        return;
    }
    switch (Format) {
    case RGD_CB:
	WriteCodebook(Filename, CB, 1);
	break;
    case RGD_TXT:
	f = fopen(Filename, "wt");
	fprintf(f, "%% %i vectors, dimensionality %i.\n",
	    BookSize(CB), VectorSize(CB));
	for (n = 0; n < BookSize(CB); n++) {
	    fprintf(f, "%% %i, freq = %i\n", n, VectorFreq(CB, n));
	    for (k = 0; k < VectorSize(CB); k++) fprintf(f, "%e ",
		Minimum + ((Maximum - Minimum) / (CB->MaxValue - CB->MinValue))
		* VectorScalar(CB, n, k));
	    fprintf(f, "\n");
	}
	fclose(f);
	break;
    default:
	ErrorMessage("cb_util.c::WriteGenericData: Unknown format.\n");
	ExitProcessing(FATAL_ERROR);
    }
}

void WriteGenericMapping(char* Filename, TRAININGSET* TS, PARTITIONING* PA,
    int Format, int AllowOverwrite)
{
    int n;
    FILE* f;
    if (!AllowOverwrite && ExistFile(Filename)) {
        ErrorMessage("File\n%s\nexists and no overwrite allowed.\n", Filename);
        return;
    }
    switch (Format) {
    case RGD_CB:
	WritePartitioning(Filename, PA, TS, 1);
	break;
    case RGD_TXT:
	f = fopen(Filename, "wt");
	fprintf(f, "%% %i vectors, %i groups.\n",
	    BookSize(TS), PartitionCount(PA));
	for (n = 0; n < BookSize(TS); n++) fprintf(f, "%i\n", Map(PA, n));
	fclose(f);
	break;
    default:
	ErrorMessage("cb_util.c::WriteGenericMapping: Unknown format.\n");
	ExitProcessing(FATAL_ERROR);
    }
}

void CopyCodebookOmitting(CODEBOOK* To, CODEBOOK* From, int Index) {
    int k;
    for (k = 0; k < BookSize(From); ++k) if (k != Index) {
        if (k < Index) CopyNode(&Node(From, k), &Node(To, k), VectorSize(To));
        else CopyNode(&Node(From, k), &Node(To, k - 1), VectorSize(To));
    }
}


int* FrequencyToIndexMap(TRAININGSET* TS) {
    int k, n, p;
    int* Map = allocate(sizeof(int) * TotalFreq(TS));
    n = 0;
    for (k = 0; k < BookSize(TS); k++) 
	for (p = 0; p < VectorFreq(TS, k); p++, n++) Map[n] = k;
    return Map;
}

int SelectNonCodevector(TRAININGSET* TS, CODEBOOK* CB, int* FreqToIndexMap) {
    int Selected, k, GetNew;
    do {
        if (FreqToIndexMap) 
            Selected = FreqToIndexMap[irand(0, TotalFreq(TS) - 1)];
        else Selected = irand(0, BookSize(TS) - 1);
        GetNew = 0;
        for (k = 0; k < BookSize(CB) && !GetNew; k++) GetNew = EqualVectors(
            Vector(CB, k), Vector(TS, Selected), VectorSize(CB));
    } while (GetNew);
    return Selected;
}

static int SwapRandom(TRAININGSET* TS,
    CODEBOOK* CB, PARTITIONING* PA,
    int* FromCluster, int* ToCluster, int* FreqToIndexMap)
{
    int Selected;
    *ToCluster = irand(0, BookSize(CB) - 1);
    Selected = SelectNonCodevector(TS, CB, FreqToIndexMap);
    CopyVector(Vector(TS, Selected), Vector(CB, *ToCluster), VectorSize(CB));
    /*VectorFreq(CB, *ToCluster) += VectorFreq(TS, Selected);
    VectorFreq(CB, Map(PA, Selected)) -= VectorFreq(TS, Selected);*/
    VectorFreq(CB, *ToCluster) = 0;
    *FromCluster = Map(PA, Selected);
    /*ChangePartition(TS, PA, *ToCluster, Selected);*/
    return *ToCluster;
}

int SwapCentroid(CriterionInfo* CI, TRAININGSET* TS,
    CODEBOOK* CB, PARTITIONING* PA, int SwapMethod,
    int* FromCluster, int* ToCluster, int* FreqToIndexMap)
{
    switch (SwapMethod) {
    case SWAP_NONE: return 0;
    case SWAP_RANDOM:
	return SwapRandom(TS, CB, PA, FromCluster, ToCluster, FreqToIndexMap);
    }
    ErrorMessage("cb_util.c:: SwapCentroid: unknown method.\n");
    ExitProcessing(FATAL_ERROR);
    return 0;
}

static void Crash(char* msg) {
    ErrorMessage(msg);
    *((int*)0) = 1;
}

void CheckCodebook(TRAININGSET* TS, CODEBOOK* CB) {
    int FSum, k, n;
    FSum = 0;
    for (k = 0; k < BookSize(CB); k++) {
	FSum += VectorFreq(CB, k);
	if (VectorFreq(CB, k) < 0) Crash("Negative frequency.\n");
	for (n = 0; n < VectorSize(CB); n++) {
	    if (VectorScalar(CB, k, n) < TS->MinValue)
		Crash("Too small value.\n");
	    if (VectorScalar(CB, k, n) > TS->MaxValue) 
		Crash("Too large value.\n");
	}
    }
    if (FSum != TotalFreq(CB)) Crash("Bad total freq.\n");
    if (FSum != TotalFreq(TS)) Crash("Inequal freqs.\n");
}

void CheckPartitioning(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA) {
    int k;
    for (k = 0; k < PA->TSsize; k++) {
	if (Map(PA, k) < 0) Crash("Map to negative.\n");
	if (Map(PA, k) > BookSize(CB)) Crash("Map too high.\n");
    }
    /*
    for (k = 0; k < PA->PartitionCount; k++) {
	if (VectorFreq(CB, k) != CCFreq(PA, k)) Crash("Different freqs.\n");
    }
    */
}

void FixCodebookFrequencies(CODEBOOK* CB, PARTITIONING* PA) {
    int k;
    for (k = 0; k< BookSize(CB); ++k) VectorFreq(CB, k) = CCFreq(PA, k);
}

