/**********************************************************

    ITER.C

    7.10.2002

    Ismo Krkkinen

    Calculation of stopping criteria for number of
    iterations.

*********************************************************/

#include "iter.h"
#include <math.h>
#include <string.h>
#include "owntypes.h"
#include "interfc.h"
#include "memctrl.h"

Iterations* itNew(I_TYPE T, int MinIterations, float Limit, int MaxIterations,
    I_StoppingFunction F)
{
    Iterations* I = allocate(sizeof(Iterations));
    if (T < I_User || I_Last <= T) {
        ErrorMessage("iter.c::itNew: unknown type\n");
	ExitProcessing(FATAL_ERROR);
    }
    I->Type = T;
    I->ValuesAllocated = 256;
    I->Values = allocate(sizeof(float) * I->ValuesAllocated);
    I->Count = 0;
    I->MinimumIterations = MinIterations;
    if (MaxIterations && MaxIterations < MinIterations)
        MaxIterations = MinIterations;
    if (T == I_Fixed) I->MaximumIterations = MinIterations;
    else I->MaximumIterations = MaxIterations;
    if (T == I_NTimes) I->Limit = 1.0f / Limit;
    else I->Limit = Limit;
    I->Func = F;
    return I;
}

void itReset(Iterations* I) {
    I->Count = 0;
}

void itDelete(Iterations* I) {
    deallocate(I->Values);
    deallocate(I);
}

static int FiftyFifty(float Limit, int Count, float* Values) {
    float Alku, Loppu;
    Alku = Values[1] - Values[Count / 2];
    Loppu = Values[Count / 2] - Values[Count - 1];
    return Alku * Limit >= Loppu;
}

static int Estimate(float Limit, int Count, float* Values) {
    /* Estimates the value 1.5 * Count by calculating how
       much less than linear improvement was when determined from
       the first half, then uses that to estimate future from the
       second half. Iterations are stopped if estimate gets too
       small compared to the improvement so far. */
    float ac, ab, bc;
    ac = Values[1] - Values[Count - 1];
    ab = Values[1] - Values[Count / 2];
    bc = Values[Count / 2] - Values[Count - 1];
    return Limit * ac * ab >= bc * bc;
}

int itStop(Iterations* I, float Value)
{
    float* tmp;
    if (I->Count == I->ValuesAllocated) {
        tmp = allocate(sizeof(float) * 2 * I->ValuesAllocated);
        memcpy(tmp, I->Values, sizeof(float) * I->ValuesAllocated);
        deallocate(I->Values);
        I->Values = tmp;
        I->ValuesAllocated *= 2;
    }
    I->Values[I->Count++] = Value;
    if (I->Count < I->MinimumIterations) return 0;
    if (I->MaximumIterations && I->Count >= I->MaximumIterations) return 1;
    switch (I->Type) {
    case I_User: return I->Func(I->Limit, I->Count, I->Values);
    case I_Fixed: return 1;
    case I_NTimes:
	return I->Values[I->Count - 1] == I->Values[(int)(I->Count * I->Limit)];
    case I_FiftyFiftyRatio:
        return FiftyFifty(I->Limit, I->Count, I->Values);
    case I_EstimatedImprovement:
        return Estimate(I->Limit, I->Count, I->Values);
    case I_Last: break;
    }
    return 1;
}

