/* ---------------------------------------------------------------------*/
/*                                                                      */
/*      matrix.cpp                                                      */
/*                                                                      */
/*    Basic matrix manipulation operations.                             */
/*                                                                      */
/*    See any numerical linear algebra book for more information        */
/*    about the definitions etc. for these operations.                  */
/*                                                                      */
/*      Copyright (C):                                                  */
/*      University of Joensuu, Department of Computer Science           */
/*                                                                      */
/*      Author(s) and current version:                                  */
/*                                                                      */
/*      Tomi Kinnunen & Ismo Krkkinen                                 */
/*      April 2002 - version 0.2                                        */
/*      tkinnu@cs.joensuu.fi, iak@cs.joensuu.fi                         */
/*                                                                      */
/*
 * Note: code in RowEchelon, Inverse and Determinant is partially copied
 * from implementation of a matrix class by Allen Barker. See
 *   Barker, String, Array, and Matrix Classes in C++
 * in:
 *   http://www.cs.virginia.edu/~alb/techrep/techrep.html
 */
/* -------------------------------------------------------------------- */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "matrix.hpp"

static int CrashVar;

static void SegFault(char* ErrorMsg) {
    fprintf(stderr, "%s\n", ErrorMsg);
    fflush(stderr);
    CrashVar = *((int*)0);
}


// From Allen Barker's code. Modified to work on *this, not on a copy.
void Matrix::RowEchelon() {
    int workCol = 0;
    for (int workRow = 0; workRow < Rows(); workRow++) {
        // find the pivot (partial pivoting strategy);
        // sum down the work column and find max abs of any element
        int pivotRow = workRow; /* Because of compiler warnings. */
        bool foundPivot = false;
        for (; workCol < Cols(); workCol++) { // go across row
            float maxAbs = 0;
            foundPivot = false;
            for (int k = workRow; k < Rows(); k++) { // go down col
                float tmpD = fabs((*this)(k, workCol));
                if (tmpD > maxAbs) {
                    maxAbs = tmpD;
                    pivotRow = k;
                }
            }
            if (maxAbs != 0) {
                foundPivot = true;
                break;
            }
        }
        if (!foundPivot) break; // done; no nonzero pivot
        float pivotVal = (*this)(pivotRow, workCol);

        // swap the workRow with the pivot row (if workRow != pivotRow)
        if (workRow != pivotRow) {
            for(int swapCol = workCol; swapCol < Cols(); swapCol++) {
                float tmpD = (*this)(pivotRow, swapCol);
                (*this)(pivotRow, swapCol) = (*this)(workRow, swapCol);
                (*this)(workRow, swapCol) = tmpD;
            }
        }

        // divide workRow by pivotVal
        for (int swapCol = workCol; swapCol < Cols(); swapCol++)
            (*this)(workRow, swapCol) /= pivotVal;

        // add to each following row workRow times the negative of 
        // the leading nonzero value in that row
        for (int k = workRow + 1; k < Rows(); k++) {
             float tmpD = (*this)(k, workCol);
             if (tmpD == 0) continue;
             for (int sumCol = workCol; sumCol < Cols(); sumCol++)
                 (*this)(k, sumCol) += -tmpD * (*this)(workRow, sumCol);
        }
    }
}

void Matrix::copy_over(const Matrix& M) {
    if (rows * cols != M.rows * M.cols) {
        delete[] Data;
        Data = new float[M.rows * M.cols];
    }
    rows = M.rows;
    cols = M.cols;
    memcpy(Data, M.Data, sizeof(float) * rows * cols);
}

Matrix::Matrix(int Rows, int Cols) : Data(0), rows(Rows), cols(Cols) {
    if (rows * cols) Data = new float[rows * cols];
}

Matrix& Matrix::operator=(const Matrix& M) {
    if (this != &M) copy_over(M);
    return *this;
}

Matrix::~Matrix() {
    if (Data) delete[] Data;
}

void Matrix::Resize(int NewRows, int NewColumns) {
    if (rows == NewRows && cols == NewColumns) return;
    Matrix M(NewRows, NewColumns);
    for (int c = 0; c < ((cols < NewColumns) ? cols : NewColumns); ++c)
        for (int r = 0; r < ((rows < NewRows) ? rows : NewRows); ++r)
            M(r, c) = (*this)(r, c);
    delete[] Data;
    Data = M.Data;
    M.Data = 0;
    rows = NewRows;
    cols = NewColumns;
}

void Matrix::SetSubMatrix(const Matrix& From, int Row, int Col) {
    if (Row < 0 || Row + From.rows > rows || Col < 0 || Col + From.cols > cols)
        SegFault("Matrix::SetSubMatrix: Region outside matrix.");
    for (int c = 0; c < From.cols; ++c) for (int r = 0; r < From.rows; ++r)
        (*this)(r + Row, c + Col) = From(r, c);
}

Matrix Matrix::SubMatrix(int Row, int Col, int RowCount, int ColCount) const {
    Matrix M(RowCount, ColCount);
    SubMatrix(M, Row, Col);
    return M;
}

void Matrix::SubMatrix(Matrix& To, int Row, int Col) const {
    if (Row < 0 || Col < 0 || Row + To.rows > rows || Col + To.cols > cols)
        SegFault("Matrix::SubMatrix: Region outside matrix.");
    float* Dest = To.Data;
    for (int r = Row; r < Row + To.Rows(); ++r)
        for (int c = Col; c < Col + To.Cols(); ++c)
            *Dest++ = (*this)(r, c);
}

void Matrix::InsertRows(const Matrix& NewRows) {
    if (NewRows.cols != cols) SegFault("Matrix::InsertRows: size mismatch.");
    Resize(rows + NewRows.rows, cols);
    SetSubMatrix(NewRows, rows - NewRows.rows, 0);
}

void Matrix::InsertColumns(const Matrix& NewColumns) {
    if (NewColumns.rows != rows)
        SegFault("Matrix::InsertColumns: size mismatch.");
    Resize(rows, cols + NewColumns.cols);
    SetSubMatrix(NewColumns, 0, cols - NewColumns.cols);
}

void Matrix::Print(FILE* F) const {
    int r, c;
    float* Curr = Data;
    for (c = 0; c < cols; ++c) {
        for (r = 0; r < rows; ++r) fprintf(F, "%g ", *Curr++);
        fprintf(F, "\n");
    }
}


void Matrix::Transpose() {
    if (rows > 1 && rows == cols) {
        float tmp;
        for (int r = 0; r < rows - 1; ++r) for (int c = r + 1; c < cols; ++c) {
            tmp = Data[r + rows * c];
            Data[r + rows * c] = Data[c + rows * r];
            Data[c + rows * r] = tmp;
        }
    } else if (rows > 1 && cols > 1) {
        float* Data2 = new float[rows * cols];
        for (int r = 0; r < rows; ++r) for (int c = 0; c < cols; ++c)
            Data2[c + cols * r] = Data[r + rows * c];
        delete[] Data;
        Data = Data2;
    }
    int s = rows;
    rows = cols;
    cols = s;
}

void Matrix::Identity() {
    float* Curr = Data;
    for (int r = 0; r < rows; ++r) for (int c = 0; c < cols; ++c)
        *Curr++ = (r == c) ? 1.0f : 0.0f;
}

void Matrix::Fill(float Value) {
    for (float* Curr = Data; Curr != Data + (rows * cols); ++Curr)
        *Curr = Value;
}

void Matrix::Randomize() {
    for (float* Curr = Data; Curr != Data + (rows * cols); ++Curr)
        *Curr = rand() / (float)RAND_MAX;
}

void Matrix::Scale(float Value) {
    for (float* Curr = Data; Curr != Data + (rows * cols); ++Curr)
        *Curr *= Value;
}

Matrix& Matrix::operator*=(const Matrix& M) {
    if (this->IsSquare() && M.IsSquare() && rows == M.rows) {
        float* Temp = new float[cols];
        for (int r = 0; r < rows; ++r) {
            for (int c = 0; c < cols; ++c) {
                Temp[c] = 0;
                for (int k = 0; k < rows; ++k)
                    Temp[c] += (*this)(r, k) * M(k, c);
            }
            memcpy(Data + cols * r, Temp, sizeof(float) * cols);
        }
        delete[] Temp;
    } else if (cols == M.rows) {
        Matrix Result(rows, M.cols);
        for (int r = 0; r < rows; ++r) for (int c = 0; c < M.cols; ++c) {
            float Sum = 0;
            for (int k = 0; k < cols; ++k) Sum += (*this)(r, k) * M(k, c);
            Result(r, c) = Sum;
        }
        delete[] Data;
        Data = Result.Data;
        Result.Data = 0;
        rows = Result.rows;
        cols = Result.cols;
    } else SegFault("Matrix::operator*=: Size mismatch.");
    return *this;
}

Matrix& Matrix::operator+=(const Matrix& M) {
    if (rows != M.rows || cols != M.cols)
        SegFault("Matrix::operator+=: size mismatch.");
    float* Curr = Data;
    float* CurrSrc = M.Data;
    for (; Curr != Data + (rows * cols); ++Curr, ++CurrSrc) *Curr += *CurrSrc;
    return *this;
}

Matrix& Matrix::operator-=(const Matrix& M) {
    if (rows != M.rows || cols != M.cols)
        SegFault("Matrix::operator-=: size mismatch.");
    float* Curr = Data;
    float* CurrSrc = M.Data;
    for (; Curr != Data + (rows * cols); ++Curr, ++CurrSrc) *Curr -= *CurrSrc;
    return *this;
}

float Matrix::Determinant() const {
    // this is just RowEchelon with more bookkeeping
    Matrix m = *this;
    float detRecord = 1;
    int workCol = 0;
    for(int workRow = 0; workRow < Rows(); workRow++) {
        // find the pivot (partial pivoting strategy);
        // sum down the work column and find max abs of any element
        int pivotRow = workRow; /* Because of compiler warnings. */
        bool foundPivot = false;
        for( ; workCol < Cols(); workCol++) { // go across row
            float maxAbs = 0;
            foundPivot = false;
            for(int k = workRow; k < Rows(); k++) { // go down col
                // take abs
                float tmpD = fabs(m(k, workCol));
                if (tmpD > maxAbs) {
                    maxAbs = tmpD;
                    pivotRow = k;
                }
            }
            if (maxAbs != 0) {
                foundPivot = true;
                break;
            }
        }
        if (!foundPivot) break; // done; no nonzero pivot
        float pivotVal = m(pivotRow, workCol);

        // swap the workRow with the pivot row (if workRow != pivotRow)
        if (workRow != pivotRow) {
            detRecord *= -1;
            for(int swapCol = workCol; swapCol < Cols(); swapCol++) {
                float tmpD = m(pivotRow, swapCol);
                m(pivotRow, swapCol) = m(workRow, swapCol);
                m(workRow, swapCol) = tmpD;
            }
        }

        // divide workRow by pivotVal
        detRecord *= pivotVal;
        for(int swapCol = workCol; swapCol < Cols(); swapCol++)
            m(workRow, swapCol) /= pivotVal;

        // add to each following row workRow times the negative of 
        // the leading nonzero value in that row
        for(int k = workRow + 1; k < Rows(); k++) {
            float tmpD = m(k, workCol);
            if (tmpD == 0) continue;
            for (int sumCol = workCol; sumCol < Cols(); sumCol++)
                m(k, sumCol) += -tmpD * m(workRow, sumCol);
        }
    }
    for (int i = 0; i < m.Rows(); i++) detRecord *= m(i, i);
    return detRecord;
}

float Matrix::Trace() const {
    float Result = 0;
    for (int r = 0; r < rows; ++r) Result += (*this)(r, r);
    return Result;
}

Matrix& Matrix::Invert() {
    if (Rows() != Cols())
        SegFault("Matrix::Invert(): matrix is not square.");
    // Left part is original matrix, right part is identity matrix. Inverse is
    // built into the right part.
    Matrix m(Rows(), 2 * Cols());
    m.SetSubMatrix(*this, 0, 0);
    for (int i = 0; i < Rows(); i++)
        for (int j = 0; j < Cols(); j++)
            if (i == j) m(i, j + Cols()) = 1;
            else m(i, j + Cols()) = 0;
    m.RowEchelon();

    // now go back up to convert the upper part of the r.h.s of
    // the augmented matrix zero
    for (int workRow = m.Rows() - 1; workRow >= 0; workRow--) {
        for (int clearRow = workRow - 1; clearRow >= 0; clearRow--) {
            float multVal = m(clearRow, workRow);
            for (int workCol = workRow; workCol < m.Cols(); workCol++)
                m(clearRow, workCol) -= m(workRow, workCol) * multVal;
        }
    }
    m.SubMatrix(*this, 0, Cols());
    return *this;
}

Matrix Matrix::Inverse() const {
    Matrix M(*this);
    M.Invert();
    return M;
}

float Matrix::ElementSum() const {
    float Sum = 0.0f;
    for (float* Curr = Data; Curr != Data + (rows * cols); ++Curr)
        Sum += *Curr;
    return Sum;
}

float Matrix::ElementProduct() const {
    float Product = 1.0f;
    for (float* Curr = Data; Curr != Data + (rows * cols); ++Curr)
        Product *= *Curr;
    return Product;
}

float Matrix::Min() const {
    float MinElem = *Data;
    for (float* Curr = Data; Curr != Data + (rows * cols); ++Curr)
        if (*Curr < MinElem) MinElem = *Curr;
    return MinElem;
}

float Matrix::Max() const {
    float MaxElem = *Data;
    for (float* Curr = Data; Curr != Data + (rows * cols); ++Curr)
        if (*Curr > MaxElem) MaxElem = *Curr;
    return MaxElem;
}

void Matrix::Diagonal(Matrix& Target) {
    int D = rows > cols ? cols : rows;
    if (D == Target.rows && Target.cols == 1
        || D == Target.cols && Target.rows == 1)
    {
        float* To = Target.Data;
        for (int k = 0; k < D; ++k) *To++ = (*this)(k, k);
    } else SegFault("Matrix::Diagonal: size mismatch.");
}

void Matrix::OuterProduct(const Matrix& A, const Matrix& B) {
#if defined(DEBUG)
    if (&A != &B) {
        if (A.Rows() * A.Cols() != B.Rows() * B.Cols())
            SegFault("Matrix::OuterProduct: matrices of different size.");
        if (!A.IsVector() || !B.IsVector())
            SegFault("Matrix::OuterProduct: matrices must be vectors.");
    } else if (!A.IsVector())
        SegFault("Matrix::OuterProduct: matrix must be vector.");
    if (A.Rows() * A.Cols() != rowsi || !IsSquare())
        SegFault("Matrix::OuterProduct: target matrix of wrong size.");
#endif
    float* Curr = Data;
    for (int r = 0; r < rows; ++r) {
        float AValue = A[r];
        for (float* BCurr = B.Data; BCurr != B.Data + rows; ++BCurr, ++Curr)
            *Curr = AValue * *BCurr;
    }
}

void Matrix::OuterProduct(ConstRowIterator& V1Begin, ConstRowIterator& V1End, 
    ConstRowIterator& V2Begin, ConstRowIterator& V2End)
{
    float* Curr = Data;
    while (V1Begin != V1End) {
        ConstRowIterator V2 = V2Begin;
        while (V2 != V2End) {
            *Curr = *V2 * *V1Begin;
            ++Curr;
            ++V2;
        }
        ++V1Begin;
    }
}

bool operator==(const Matrix& A, const Matrix& B) {
    if (A.Rows() != B.Rows() || A.Cols() != B.Cols()) return false;
    if (&A == &B) return true;
    float* APtr = A.Data;
    float* BPtr = B.Data;
    const float Tolerance = 1e-6;
    for (; APtr != A.Data + (A.rows * A.cols); ++APtr, ++BPtr) {
        float AVal = fabs(*APtr);
        float BVal = fabs(*BPtr);
        if (fabs(*APtr - *BPtr) / ((AVal >= BVal) ? AVal : BVal) > Tolerance)
            return false;
    }
    return true;
}

Matrix Transpose(const Matrix& M) {
    Matrix A(M);
    A.Transpose();
    return A;
}

Matrix operator*(const Matrix& M, float Scalar) {
    Matrix A(M);
    A.Scale(Scalar);
    return A;
}

Matrix operator*(float Scalar, const Matrix& M) {
    return M * Scalar;
}

Matrix operator*(const Matrix& A, const Matrix& B) {
    Matrix M(A);
    M *= B;
    return M;
}

Matrix operator+(const Matrix& A, const Matrix& B) {
    Matrix M(A);
    M += B;
    return M;
}

Matrix operator-(const Matrix& A, const Matrix& B) {
    Matrix M(A);
    M -= B;
    return M;
}

float Dot(const Matrix& A, const Matrix& B) {
    if ((A.rows == 1 || A.cols == 1) && (B.rows == 1 || B.cols == 1)
        && (A.rows * A.cols == B.rows * B.cols))
    {
        float* APtr = A.Data;
        float* BPtr = B.Data;
        float Sum = 0;
        for (; APtr != A.Data + (A.rows * A.cols); ++APtr, ++BPtr)
            Sum += *APtr * *BPtr;
        return Sum;
    } else SegFault("matrix.cpp::Dot: Size mismatch.");
    return 0.0f; // Appease compiler.
}



/*
 * Old code.
 */



#if 0
MATRIX TriL(MATRIX M) {
/* Lower triangular of M. Result is given in a column vector. */

    return void;

}
#endif

MATRIX TriU(MATRIX M) {
    int i, j, k, NumElems;
    MATRIX Result;
    /* You can verify this formula with pen and pencil if really want: */
    NumElems =(int)(MRows(M) * MCols(M) - 0.5 * MRows(M) *(MRows(M) + 1));
    Result = NewMatrix(NumElems, 1);
    k = 1;
    for (i = 1; i <= MRows(M); i++) {
        for (j = i + 1; j <= MCols(M); j++) {
            SetMatrixElement(Result, k, 1, GetMatrixElement(M, i, j));
            k++;
        }
    }
    return Result;
}

/*
 * End of old code.
 */



int MRows(MATRIX M) {
    return M->Rows();
}

int MCols(MATRIX M) {
    return M->Cols();
}

int IsSquare(MATRIX M) {
    return M->IsSquare();
}

int IsScalar(MATRIX M) {
    return M->IsScalar();
}

int IsColVector(MATRIX M) {
    return M->IsColVector();
}

int IsRowVector(MATRIX M) {
    return M->IsRowVector();
}

MATRIX NewMatrix(int Rows, int Cols) {
    return new Matrix(Rows, Cols);
}

void NewMatrixAllocated(MATRIX M, int Rows, int Cols) {
    memset(M, 0, sizeof(Matrix));
    M->Resize(Rows, Cols);
}

void DeleteMatrix(MATRIX M) {
    delete M;
}

void DeleteMatrixAllocated(MATRIX M) {
    if (!M) return;
    M->~Matrix();
    memset(M, 0, sizeof(Matrix));
}

void CopyMatrix(MATRIX From, MATRIX To) {
    *To = *From;
}

void PrintMatrix(MATRIX M, FILE * f) {
    M->Print(f);
}

void SetMatrixElement(MATRIX M, int Row, int Col, float Value) {
   (*M)(Row - 1, Col - 1) = Value;
}

float GetMatrixElement(MATRIX M, int Row, int Col) {
    return (*M)(Row - 1, Col - 1);
}

MATRIX GetRow(MATRIX TargetVector, MATRIX M, int Row) {
    if ((Row < 1) ||(Row > MRows(M)))
        SegFault("ERROR [GetRow]: Matrix subscript out of range!");
    else if (!(IsRowVector(TargetVector)))
        SegFault("ERROR [GetRow]: Target vector is not row vector!");
    M->SubMatrix(*TargetVector, Row - 1, 0);
    return TargetVector;
}

MATRIX GetColumn(MATRIX TargetVector, MATRIX M, int Col) {
    if ((Col < 1) ||(Col > MCols(M)))
        SegFault("ERROR [GetColumn]: Matrix subscript out of range!");
    else if (!(IsColVector(TargetVector)))
        SegFault("ERROR [GetColumn]: Target vector is not column vector!");
    M->SubMatrix(*TargetVector, 0, Col - 1);
    return TargetVector;
}

MATRIX SetRow(MATRIX M, MATRIX Source, int Row) {
    if ((Row < 1) ||(Row > MRows(M)))
        SegFault("ERROR [SetRow]: Matrix subscript out of range!");
    else if (MCols(Source) != MCols(M))
        SegFault("ERROR [SetRow]: Mismatched number of columns in matrices!");
    else if (MRows(Source) > 1)
        SegFault("ERROR [SetRow]: Matrix must have only one row!");
    M->SetSubMatrix(*Source, Row - 1, 0);
    return M;
}

MATRIX SetColumn(MATRIX M, MATRIX Source, int Col) {
    if ((Col < 1) ||(Col > MCols(M)))
        SegFault("ERROR [SetColumn]: Matrix subscript out of range!");
    else if (MRows(Source) != MRows(M))
        SegFault("ERROR [SetColumn]: Mismatched number of rows in matrices!");
    else if (MCols(Source) > 1)
        SegFault("ERROR [SetColumn]: Matrix must have only one column!");
    M->SetSubMatrix(*Source, 0, Col - 1);
    return M;
}


MATRIX InsertNewVectorToMatrix(MATRIX M, MATRIX Vector, int ColVec) {
/* ColVec != 0 --> insert a column vector. Otherwise row vector */
    if (ColVec && (MRows(Vector) != MRows(M)))
        SegFault("ERROR [InsertNewVector]: Mismatched number of rows!");
    if (!ColVec && (MCols(Vector) != MCols(M)))
        SegFault("ERROR [InsertNewVector]: Mismatched number of columns!");
    if (!ColVec) M->InsertRows(*Vector);
    else M->InsertColumns(*Vector);
    return M;
}

MATRIX Diag(MATRIX TargetVector, MATRIX M) {
    if (MRows(TargetVector) != MRows(M))
        SegFault("ERROR [DiagonalElements] : Dimension mismatch.");
    M->Diagonal(*TargetVector);
    return TargetVector;
}

int Same(MATRIX A, MATRIX B) {
    return *A == *B;
}

MATRIX Transpose(MATRIX M) {
    M->Transpose();
    return M;
}

MATRIX IdentityMatrix(MATRIX M) {
    M->Identity();
    return M;
}

MATRIX ZeroMatrix(MATRIX M) {
    M->Fill(0);
    return M;
}

MATRIX OneMatrix(MATRIX M) {
    M->Fill(1);
    return M;
}

MATRIX FillMatrix(MATRIX M, float Value) {
    M->Fill(Value);
    return M;
}

MATRIX RandomMatrix(MATRIX M) {
    M->Randomize();
    return M;
}

MATRIX MultiplyByScalar(MATRIX Result, MATRIX M, float Scalar) {
    *Result = *M;
    Result->Scale(Scalar);
    return Result;
}

MATRIX MatrixSum(MATRIX Result, MATRIX A, MATRIX B) {
    *Result = *A;
    *Result += *B;
    return Result;
}

MATRIX MatrixSubstract(MATRIX Result, MATRIX A, MATRIX B) {
    *Result = *A;
    *Result -= *B;
    return Result;
}

MATRIX MatrixProduct(MATRIX Result, MATRIX A, MATRIX B) {
    *Result = *A * *B;
    return Result;
}

float Determinant(MATRIX M) {
    return M->Determinant();
}

float Trace(MATRIX A) {
    return A->Trace();
}

MATRIX MatrixInverse(MATRIX Result, MATRIX A) {
    *Result = *A;
    Result->Invert();
    return Result;
}

float ProdOfElements(MATRIX M) {
    return M->ElementProduct();
}

float SumOfElements(MATRIX M) {
    return M->ElementSum();
}

float MaxElem(MATRIX M) {
    return M->Max();
}

float MinElem(MATRIX M) {
    return M->Min();
}

float VectorDistanceM(MATRIX x, MATRIX y, int p) {
    int i;
    float D = 0;
    if (!IsScalar(x) &&(IsRowVector(x) || IsRowVector(y)))
        SegFault("ERROR [VectorDistance] : Use column vectors, please!");
    if (MRows(x) != MRows(y))
        SegFault("ERROR [VectorDistance] : dimension mismatch!");
    if (p <= 0)
        SegFault("ERROR [VectorDistance] : p should be > 0");
    for (i = 0; i < MRows(x); i++)
        D += pow(fabs((*x)(i, 0) - (*y)(i, 0)), p);
    return pow(D, 1.0 / p);
}

float VectorMaxDistance(MATRIX x, MATRIX y) {
    int i;
    float MaxDiff = -1.0;
    for (i = 0; i < MRows(x); i++) {
        float Diff = fabs((*x)(i, 0) - (*y)(i, 0));
        if (Diff > MaxDiff) MaxDiff = Diff;
    }
    return MaxDiff;
}

float Avg(MATRIX Vector) {
/* Average value of Vector */
    if (!IsColVector(Vector) && !IsRowVector(Vector))
         SegFault("ERROR[Avg]: You should give a vector as input.");
    return Vector->ElementSum() / (MRows(Vector)* MCols(Vector));
}

float Var(MATRIX Vector, int UnbiasedEstimate) {
/* Variance of the random variables given in the Vector.

   Vector is type(N x 1).
   UnbiasedEstimate != 0:
    Normalize by N-1
   Else
    Normalize by N.
*/
    int NumElem;
    float Mean, SqSum = 0;
    if (!IsColVector(Vector) && !IsRowVector(Vector))
        SegFault("ERROR[Var]: You should give a vector as input.");
    Mean = Avg(Vector);
    NumElem = MRows(Vector) * MCols(Vector);
    for (int c = 0; c < MCols(Vector); ++c)
        for (int r = 0; r < MRows(Vector); ++r)
            SqSum += ((*Vector)(r, c) - Mean) * ((*Vector)(r, c) - Mean);
    if (UnbiasedEstimate && NumElem > 1) --NumElem;
    return SqSum / NumElem;
}

float Norm(MATRIX Vector) {
    if (!IsColVector(Vector)) SegFault("ERROR [Norm] : give a col vector.\n");
    return sqrt(Dot(*Vector, *Vector));
}

float InnerProduct(MATRIX Vec1, MATRIX Vec2) {
    if (!IsRowVector(Vec1) && !IsColVector(Vec2)) SegFault(
        "[InnerProduct]: inputs should be : Vec1 =(1 x d), Vec2 =(d x 1).");
    return Dot(*Vec1, *Vec2);
}

float VectorSimilarity(MATRIX Vec1, MATRIX Vec2,
    VECTOR_SIMILARITY_TYPE SimType, float Alpha)
{
    float dist, Val = 0;
    Matrix Diff;
    if (MCols(Vec1) > 1 || MCols(Vec2) > 1)
        SegFault("ERROR [VectorSimilarity]: Column vector expected.");
    else if (MRows(Vec1) != MRows(Vec2))
        SegFault("ERROR [VectorSimilarity]: Dimension mismatch of vectors.");
    else if (Alpha <= 0)
        SegFault("ERROR [VectorSimilarity]: Alpha must be positive.");
    switch (SimType) {
    case smooth_Gauss:
        Diff = *Vec1 - *Vec2;
        dist = Norm(&Diff);    // Euclidean distance
        /* Normalize squared distance by dimensionality */
        dist = dist * dist / MRows(Vec1);
        Val = (1.0 / Alpha)
            * ((exp(-0.5 *(pow((dist / 1.0 * Alpha), 2)))) - 1) + 1;
        break;
    case Lp_similarity:
        Diff = *Vec1 - *Vec2;
        dist = Dot(Diff, Diff);
        Val = pow((1 + dist), -2.0);
        break;
    }
    return Val;
}

void NormalizeSumToUnity(MATRIX M) {
/* After normalization, sum of elements of M is 1. */
    float Sum = M->ElementSum();
    if (Sum) M->Scale(1.0f / Sum);
}


