#if !defined(MATRIX_HPP)
#define MATRIX_HPP

#include <stdio.h>

#if defined(__cplusplus)

/*
 * Note: all iterators will be invalid if the size of the matrix changes.
 */

typedef float *RowIterator;
typedef float const *ConstRowIterator;

// Allows contents to be changed.
class ColumnIterator {
private:
    float* pos;
    int step;

public:
    ColumnIterator(float* Pos, int Step) : pos(Pos), step(Step) { }

    inline ColumnIterator& operator++(int) { pos += step; return *this; }
    inline ColumnIterator operator++() {
        ColumnIterator Orig = *this;
        pos += step;
        return Orig;
    }

    inline float& operator*() { return *pos; }

    inline bool operator==(const ColumnIterator& CI) const
        { return pos == CI.pos; }
    inline bool operator!=(const ColumnIterator& CI) const
        { return pos != CI.pos; }
};

// Does not allow contents to be changed.
class ConstColumnIterator {
private:
    float const* pos;
    int step;

public:
    ConstColumnIterator(float const* Pos, int Step) : pos(Pos), step(Step) { }

    inline ConstColumnIterator& operator++(int) { pos += step; return *this; }
    inline ConstColumnIterator operator++() {
        ConstColumnIterator Orig = *this;
        pos += step;
        return Orig;
    }

    inline float operator*() const { return *pos; }

    inline bool operator==(const ConstColumnIterator& CI) const
        { return pos == CI.pos; }
    inline bool operator!=(const ConstColumnIterator& CI) const
        { return pos != CI.pos; }
};

class Matrix {
private:
    float* Data;
    int rows, cols;

    void RowEchelon();
    void copy_over(const Matrix& M);

public:
    Matrix(int Rows = 0, int Cols = 0);
    Matrix(const Matrix& M) : Data(0), rows(0), cols(0) { copy_over(M); }
    Matrix& operator=(const Matrix& M);
    ~Matrix();

    void Resize(int NewRows, int NewColumns);
    void SetSubMatrix(const Matrix& From, int Row, int Col);
    Matrix SubMatrix(int Row, int Col, int RowCount, int ColCount) const;
    void SubMatrix(Matrix& To, int Row, int Col) const;
    // Note: sequential insertions are more efficient using one Resize
    // and several SetSubMatrix-calls.
    void InsertRows(const Matrix& NewRows);
    void InsertColumns(const Matrix& NewColumns);

    inline int Rows() const { return rows; }
    inline int Cols() const { return cols; }
    inline bool IsSquare() const { return rows == cols; }
    inline bool IsScalar() const { return rows == 1 && cols == 1; }
    inline bool IsColVector() const { return cols == 1; }
    inline bool IsRowVector() const { return rows == 1; }
    inline bool IsVector() const { return rows == 1 || cols == 1; }

    void Print(FILE* F) const;

    inline float operator()(int Row, int Col) const
        { return Data[Row * cols + Col]; }
    float& operator()(int Row, int Col) { return Data[Row * cols + Col]; }

    // These treat data as one-dimensional array.
    inline float operator[](int Index) const { return Data[Index]; }
    inline float& operator[](int Index) { return Data[Index]; }

    // So far only forward iterators provided. Add the rest if needed.
    RowIterator RowBegin(int Row) { return Data + Row * Cols(); }
    RowIterator RowEnd(int Row) { return Data + (Row + 1) * Cols(); }
    ConstRowIterator ConstRowBegin(int Row) const
        { return Data + Row * Cols(); }
    ConstRowIterator ConstRowEnd(int Row) const
        { return Data + (Row + 1) * Cols(); }

    inline ColumnIterator ColumnBegin(int Column)
        { return ColumnIterator(Data + Column, cols); }
    inline ColumnIterator ColumnEnd(int Column)
        { return ColumnIterator(Data + rows * cols + Column, cols); }
    inline ConstColumnIterator ConstColumnBegin(int Column) const
        { return ConstColumnIterator(Data + Column, cols); }
    inline ConstColumnIterator ConstColumnEnd(int Column) const
        { return ConstColumnIterator(Data + rows * cols + Column, cols); }

    void Transpose(); // In-place.
    void Identity();
    void Fill(float Value);
    void Randomize();
    void Scale(float Value);

    Matrix& operator*=(float Scalar) { this->Scale(Scalar); return *this; }
    Matrix& operator*=(const Matrix& M);
    Matrix& operator+=(const Matrix& M);
    Matrix& operator-=(const Matrix& M);

    float Determinant() const;
    float Trace() const;
    Matrix& Invert(); // Operates on this.
    Matrix Inverse() const; // Generates new matrix.
    float ElementSum() const;
    float ElementProduct() const;
    float Min() const;
    float Max() const;
    void Diagonal(Matrix& Target);
    // Computes the outer product of two vectors, it does not matter if they
    // are row or column vectors.
    void OuterProduct(const Matrix& A, const Matrix& B);
    // Outer product of two row vectors, performs no checks. Both vectors can
    // be the same.
    void OuterProduct(ConstRowIterator& V1Begin, ConstRowIterator& V1End,
        ConstRowIterator& V2Begin, ConstRowIterator& V2End);

    friend bool operator==(const Matrix& A, const Matrix& B);
    friend bool operator!=(const Matrix& A, const Matrix& B);
    friend float Dot(const Matrix& A, const Matrix& B);
};

typedef Matrix *MATRIX;

bool operator==(const Matrix& A, const Matrix& B);
inline bool operator!=(const Matrix& A, const Matrix& B) { return !(A == B); }
Matrix Transpose(const Matrix& M); // Creates new matrix.
Matrix operator*(const Matrix& M, float Scalar);
Matrix operator*(float Scalar, const Matrix& M);
Matrix operator*(const Matrix& A, const Matrix& B);
Matrix operator+(const Matrix& A, const Matrix& B);
Matrix operator-(const Matrix& A, const Matrix& B);
float Dot(const Matrix& A, const Matrix& B);

template<class Iterator, class ConstIterator>
void Difference(Iterator& Result, const Iterator& ResultEnd,
    ConstIterator& Vec1, ConstIterator& Vec2)
{
    while (Result != ResultEnd) {
        *Result = *Vec1 - *Vec2;
        ++Result;
        ++Vec1;
        ++Vec2;
    }
}

template<class Iterator, class ConstIterator>
void Sum(Iterator& Result, const Iterator& ResultEnd,
    ConstIterator& Vec1, ConstIterator& Vec2)
{
    while (Result != ResultEnd) {
        *Result = *Vec1 + *Vec2;
        ++Result;
        ++Vec1;
        ++Vec2;
    }
}

template<class Iterator>
float DistanceSquared(Iterator& Vec1, const Iterator& Vec1End, Iterator& Vec2)
{
    float Sum = 0;
    while (Vec1 != Vec1End) {
        float Diff = *Vec1 - *Vec2;
        Sum += Diff * Diff;
        ++Vec1;
        ++Vec2;
    }
    return Sum;
}

#else /* Type definitions for plain C. */


typedef struct {
   float*  Data;
   int	   Rows;
   int	   Cols;
} MATRIX_TYPE;

typedef MATRIX_TYPE* MATRIX;


#endif /* __cplusplus */


typedef enum {
   smooth_Gauss,
   Lp_similarity
} VECTOR_SIMILARITY_TYPE;


/*
 * Compatibility code.
 */

#if defined(__cplusplus)
extern "C" {
#endif

int MRows(MATRIX M);
int MCols(MATRIX M);
int IsSquare(MATRIX M);
int IsScalar(MATRIX M);
int IsColVector(MATRIX M);
int IsRowVector(MATRIX M);

/* Memory control -------------------------------------------------- */
MATRIX	NewMatrix(int Rows, int Cols);
void    NewMatrixAllocated(MATRIX M, int Rows, int Cols);
void	DeleteMatrix(MATRIX M);
void    DeleteMatrixAllocated(MATRIX M);
void 	CopyMatrix(MATRIX From, MATRIX To);

/* Matrix I/O ------------------------------------------------------ */
void    PrintMatrix(MATRIX M, FILE* f);
void	SetMatrixElement(MATRIX M, int Row, int Col, float Value);
float	GetMatrixElement(MATRIX M, int Row, int Col);

/* These return actual copies to work with the old code. */
MATRIX	GetRow(MATRIX TargetVector, MATRIX M, int Row);
MATRIX  GetColumn(MATRIX TargetVector, MATRIX M, int Col);
MATRIX	SetRow(MATRIX M, MATRIX Source, int Row);
MATRIX  SetColumn(MATRIX M, MATRIX Source, int Col);

MATRIX  InsertNewVectorToMatrix(MATRIX M, MATRIX Vector, int ColVec);
MATRIX	Diag(MATRIX TargetVector, MATRIX M);
MATRIX  TriL(MATRIX M);
MATRIX  TriU(MATRIX M);

/* Matrix operations ----------------------------------------------- */
int	Same(MATRIX A, MATRIX B);
MATRIX	Transpose(MATRIX M);
MATRIX	IdentityMatrix(MATRIX M);
MATRIX	ZeroMatrix(MATRIX M);
MATRIX  OneMatrix(MATRIX M);
MATRIX	FillMatrix(MATRIX M, float Value);
MATRIX	RandomMatrix(MATRIX M);
MATRIX	MultiplyByScalar(MATRIX Result, MATRIX M, float Scalar);
MATRIX	MatrixSum(MATRIX Result, MATRIX A, MATRIX B);
MATRIX	MatrixSubstract(MATRIX Result, MATRIX A, MATRIX B);
MATRIX	MatrixProduct(MATRIX Result, MATRIX A, MATRIX B);
float	Determinant(MATRIX M);
float	Trace(MATRIX A);
MATRIX  MatrixInverse(MATRIX Result, MATRIX A);
float   ProdOfElements(MATRIX M);
float   SumOfElements(MATRIX M);
float   MaxElem(MATRIX M);
float   MinElem(MATRIX M);

/* Vector operations ----------------------------------------------- */
float 	VectorDistanceM(MATRIX x, MATRIX y, int p);
float	VectorMaxDistance(MATRIX x, MATRIX y);
float 	Avg(MATRIX Vector);
float	Var(MATRIX Vector, int UnbiasedEstimate);
float	Norm(MATRIX Vector);
float	InnerProduct(MATRIX Vec1, MATRIX Vec2);
float	VectorSimilarity( MATRIX Vec1, MATRIX Vec2, 
			  VECTOR_SIMILARITY_TYPE SimType,
			  float Alpha );
void	NormalizeSumToUnity(MATRIX x);

#if defined(__cplusplus)
}
#endif

#endif

