/******************************************************************

    TRACEDB.C

    20.6.2000

    Ismo Krkkinen

    Routines for handling a "database" consisting of repeat, number
    of clusters, iteration and value of criterion.

******************************************************************/

#include "tracedb.h"
#include "interfc.h"
#include <limits.h>

TDBRanges* tdbrNew(int VectorCount) {
    TDBRanges* R = allocate(sizeof(TDBRanges));
    tdbrInitialize(R, VectorCount);
    return R;
}

void tdbrInitialize(TDBRanges* R, int VectorCount) {
    int k;
    for (k = 0; k < TDB_DIMENSIONALITY + 1; k++) {
	R->Begin[k] = 0;
	R->End[k] = VectorCount;
    }
}

void tdbrDelete(TDBRanges* R) {
    deallocate(R);
}

int tdbRead(float*** Set, int* Count, FILE* f) {
    int Dimensionality;
    if (!fvReadSet(Set, Count, &Dimensionality, f)) return 1;
    if (Dimensionality != TDB_DIMENSIONALITY) {
	fvDeleteSet(*Set, *Count);
	return 2;
    }
    fvSortSet(*Set, *Count, Dimensionality, 0, 0, fvAscendingOrder);
    return 0;
}

int tdbSeek(float** Set, TDBRanges* R, int Dimension, float Value) {
    int Delta, Location, RDim;
    if (Dimension >= TDB_DIMENSIONALITY) {
	ErrorMessage("tracedb.c::tdbSeek: Only %i dimensions allowed.\n",
	    TDB_DIMENSIONALITY);
	return 0;
    }
    if (!Dimension) RDim = TDB_DIMENSIONALITY;
    else RDim = Dimension - 1;
    if (R->Begin[RDim] == R->End[RDim]) {
	R->Begin[Dimension] = R->End[Dimension] = R->Begin[RDim];
	return 0;
    }
    if (Set[R->End[RDim] - 1][Dimension] < Value) {
	R->Begin[Dimension] = R->End[Dimension] = R->End[RDim] - 1;
	return 0;
    }
    if (Set[R->Begin[RDim]][Dimension] > Value) {
	R->Begin[Dimension] = R->End[Dimension] = R->Begin[RDim];
	return 0;
    }
    Location = R->Begin[RDim] + (R->End[RDim] - R->Begin[RDim]) / 2;
    Delta = (R->End[RDim] - R->Begin[RDim]) / 4;
    do {
	if (Set[Location][Dimension] < Value) Location += Delta;
	else if (Set[Location][Dimension] > Value) Location -= Delta;
	Delta /= 2;
	if (Delta < 1) Delta = 1;
	if (Set[Location][Dimension] < Value 
	    && Set[Location + 1][Dimension] > Value) break;
    } while (Set[Location][Dimension] != Value);
    while (Location > R->Begin[RDim] && Set[Location - 1][Dimension] == Value)
	Location--;
    R->Begin[Dimension] = Location;
    if (Set[Location][Dimension] != Value) {
	R->End[Dimension] = Location;
	return 0;
    }
    while (Location < R->End[RDim] && Set[Location][Dimension] == Value)
	Location++;
    R->End[Dimension] = Location;
    return R->End[Dimension] - R->Begin[Dimension];
}

int tdbFind(float** Set, TDBRanges* R,
    float Val0, float Val1, float Val2, float Val3)
{
    tdbSeek(Set, R, 0, Val0);
    tdbSeek(Set, R, 1, Val1);
    tdbSeek(Set, R, 2, Val2);
    return tdbSeek(Set, R, 3, Val3);
}

static int float_compare(void* a, void* b) {
    float aa, bb;
    aa = *(float*)a;
    bb = *(float*)b;
    if (aa < bb) return -1;
    if (aa > bb) return 1;
    return 0;
}

float** tdbValueMAMM(float** Set, int Count, int* MAMMCount) {
    TDBRanges R;
    int Total, k, n;
    float Sum;
    float* Repeats;
    int RepeatCount, ValueCount;
    float* Values;
    float** MAMM = tdbCombine(Set, Count, 4, MAMMCount);
    Total = *MAMMCount / 4;
    Repeats = fvSetUnique(Set, Count, 0, &RepeatCount);
    Values = allocate(sizeof(float) * RepeatCount);
    tdbrInitialize(&R, Count);
    for (k = 0; k < Total; k++) {
	ValueCount = 0;
	Sum = 0.0;
	for (n = 0; n < RepeatCount; n++) {
	    tdbFind(Set, &R, Repeats[n], MAMM[k][1], MAMM[k][2], 0);
	    if (tdbrBegin(&R, 1) != tdbrEnd(&R, 1))
	        Sum += Values[ValueCount++] = Set[tdbrBegin(&R, 2)][3];
	}
	fvSort(Values, ValueCount, 1);
	MAMM[k + Total * TDB_MAMM_MIN][3] = Values[0];
	MAMM[k + Total * TDB_MAMM_AVG][3] = Sum / ValueCount;
	MAMM[k + Total * TDB_MAMM_MED][3] = Values[ValueCount / 2];
	MAMM[k + Total * TDB_MAMM_MAX][3] = Values[ValueCount - 1];
    }
    deallocate(Values);
    deallocate(Repeats);
    return MAMM;
}

float** tdbIterationMAMM(float** Set, int Count, int* MAMMCount) {
    float** MAMM;
    float* Ms;
    float* Rs;
    int RCount;
    float* tmp;
    TDBRanges R;
    int ValueCount, k, n;
    float Sum;
    Rs = fvSetUnique(Set, Count, 0, &RCount);
    tmp = allocate(sizeof(float) * RCount);
    Ms = fvSetUnique(Set, Count, 1, MAMMCount);
    MAMM = fvNewSet(*MAMMCount, 5);
    tdbrInitialize(&R, Count);
    for (n = 0; n < *MAMMCount; n++) {
	ValueCount = 0;
	Sum = 0.0;
        for (k = 0; k < RCount; k++) {
	    tdbFind(Set, &R, Rs[k], Ms[n], INT_MAX, 0);
	    if (tdbrBegin(&R, 1) != tdbrEnd(&R, 1))
		Sum += tmp[ValueCount++] = Set[tdbrBegin(&R, 2)][2];
	}
	fvSort(tmp, ValueCount, 1);
	MAMM[n][0] = Ms[n];
	MAMM[n][TDB_MAMM_MIN + 1] = tmp[0];
	MAMM[n][TDB_MAMM_AVG + 1] = Sum / ValueCount;
	MAMM[n][TDB_MAMM_MED + 1] = tmp[ValueCount / 2];
	MAMM[n][TDB_MAMM_MAX + 1] = tmp[ValueCount - 1];
    }
    deallocate(tmp);
    deallocate(Ms);
    deallocate(Rs);
    return MAMM;
}

float** tdbCombine(float** Set, int Count, int Times, int* CombinedCount) {
    float* Groups, *Iters;
    int GroupCount, IterCount;
    float** Result;
    int g, t, it, k;
    Groups = fvSetUnique(Set, Count, 1, &GroupCount);
    Iters = fvSetUnique(Set, Count, 2, &IterCount);
    *CombinedCount = Times * GroupCount * IterCount;
    Result = fvNewSet(*CombinedCount, 4);
    for (t = 0; t < Times; t++)
	for (g = 0; g < GroupCount; g++)
	    for (it = 0; it < IterCount; it++)
    {
	k = t * GroupCount * IterCount + g * IterCount + it;
	Result[k][0] = t;
	Result[k][1] = Groups[g];
	Result[k][2] = Iters[it];
	Result[k][3] = 0.0;
    }
    deallocate(Groups);
    deallocate(Iters);
    return Result;
}
