#include <stdio.h>
#include "vptree.h"
#include "memctrl.h"

static int vptnAsc(const void* A, const void* B) {
    const VPTreeNode* a = (VPTreeNode*)A;
    const VPTreeNode* b = (VPTreeNode*)B;
    if (a->Dist > b->Dist) return 1;
    if (a->Dist < b->Dist) return -1;
    return 0;
}

static void MakeTree(TRAININGSET* TS, DistanceInfo* DI,
    VPTreeNode* Root, int Index, int Count, VPTreeNode* Array)
{
    int k, MaxIndex;
    float MaxDist, Dist;
    if (Count == 1) {
        /*if (Root[Index].Point != -1) *((int*)0) = 1;*/
        Root[Index].Point = Array->Point;
        Root[Index].Dist = -1.0f;
    } else {
        /* Get reasonable vantage point. */
        MaxDist = 0;
        MaxIndex = 0;
        for (k = 0; k < Count; ++k) {
            Dist = diDistance(
                DI, TS, 0, 0, Array[0].Point, Array[k].Point, DT_DATA2DATA);
            if (Dist > MaxDist) {
                MaxDist = Dist;
                MaxIndex = Array[k].Point;
            }
        }
        /* Arrange points in order from the vantage point. */
        for (k = 0; k < Count; ++k) Array[k].Dist = diDistance(
            DI, TS, 0, 0, MaxIndex, Array[k].Point, DT_DATA2DATA);
        qsort(Array, Count, sizeof(VPTreeNode), &vptnAsc);
        Root[Index].Point = MaxIndex;
        Root[Index].Dist = Array[Count / 2].Dist;
        MakeTree(TS, DI, Root, Index * 2 + 1, Count / 2, Array);
        MakeTree(TS, DI, Root, Index * 2 + 2, Count - (Count / 2),
            Array + Count / 2);
    }
}

VPTree* vptNew(TRAININGSET* TS, DistanceInfo* DI) {
    int k;
    VPTreeNode* Array;
    VPTree* V = allocate(sizeof(VPTree));
    Array = allocate(sizeof(VPTreeNode) * BookSize(TS));
    V->TS = TS;
    V->DI = DI;
    V->BufferSize = 0;
    V->Buffer = 0;
    V->Count = 1;
    V->PointsTouched = 0;
    V->SearchCount = 0;
    while (V->Count < 2 * BookSize(TS)) V->Count *= 2;
    V->Tree = allocate(sizeof(VPTreeNode) * V->Count);
    /*for (k = 0; k < V->Count; ++k) V->Tree[k].Point = -1;*/
    for (k = 0; k < BookSize(TS); ++k) Array[k].Point = k;
    MakeTree(TS, DI, V->Tree, 0, BookSize(TS), Array);
    deallocate(Array);
    return V;
}

void vptDelete(VPTree* V) {
    if (!V) return;
    fprintf(stderr,
        "Searches touched %i points (%.3f), N*N/2 is %i (%i).\n",
        V->PointsTouched, V->PointsTouched / (float)V->SearchCount,
        BookSize(V->TS) * BookSize(V->TS) / 2, BookSize(V->TS));
    deallocate(V->Tree);
    deallocate(V->Buffer);
    deallocate(V);
}

static int Branch(VPTree* V, int Point, int Found, int Count, int RootIndex) {
    float Dist, td;
    int k, tp;
    VPTreeNode* Root = V->Tree + RootIndex;
    Dist = diDistance(V->DI, V->TS, 0, 0, Root->Point, Point, DT_DATA2DATA);
    if (Root->Dist < 0) {
        V->PointsTouched++;
        V->Buffer[Found + 1].Point = Root->Point;
        V->Buffer[Found + 1].Dist = Dist;
        k = Found;
        /* Note: first element of array has distance 0. */
        while (Dist < V->Buffer[k].Dist) {
            td = V->Buffer[k].Dist;
            tp = V->Buffer[k].Point;
            V->Buffer[k].Dist = V->Buffer[k + 1].Dist;
            V->Buffer[k].Point = V->Buffer[k + 1].Point;
            V->Buffer[k + 1].Dist = td;
            V->Buffer[k + 1].Point = tp;
            --k;
        }
        if (Found < Count) return Found + 1;
    } else {
        /* Does inner sphere intersect with point surroundings? */
        if (Found < Count || Dist <= V->Buffer[Found].Dist + Root->Dist)
            Found = Branch(V, Point, Found, Count, RootIndex * 2 + 1);
        /* Does outer shell intersect with point surroundings? */
        if (Found < Count || Dist + V->Buffer[Found].Dist >= Root->Dist)
            Found = Branch(V, Point, Found, Count, RootIndex * 2 + 2);
    }
    return Found;
}

static int SeekNode(VPTree* V, int Point, int Found, int Count, int RootIndex) {
    float Dist;
    VPTreeNode* Root = V->Tree + RootIndex;
    if (Root->Dist < 0) return 0;
    Dist = diDistance(V->DI, V->TS, 0, 0, Root->Point, Point, DT_DATA2DATA);
    if (Dist < Root->Dist) {
        Found = SeekNode(V, Point, Found, Count, RootIndex * 2 + 1);
        Found = Branch(V, Point, Found, Count, RootIndex * 2 + 2);
    } else {
        Found = SeekNode(V, Point, Found, Count, RootIndex * 2 + 2);
        Found = Branch(V, Point, Found, Count, RootIndex * 2 + 1);
    }
    return Found;
}

void vptNeighbors(VPTree* V, int Point,
    int Count, int* Indices, float* Distances)
{
    int k;
    if (Count < 1) return;
    if (Count >= BookSize(V->TS)) Count = BookSize(V->TS) - 1;
    if (V->BufferSize < Count + 2) {
        deallocate(V->Buffer);
        V->Buffer = allocate(sizeof(VPTreeNode) * (Count + 2));
        V->BufferSize = Count + 2;
    }
    V->Buffer[0].Dist = 0;
    V->PointsTouched++;
    V->SearchCount++;
    SeekNode(V, Point, 0, Count, 0);
    for (k = 0; k < Count; ++k) {
        if (Indices) Indices[k] = V->Buffer[k + 1].Point;
        if (Distances) Distances[k] = V->Buffer[k + 1].Dist;
    }
}

