/******************************************************************

    CBFCM.C

    Ismo Krkkinen

    Randomized FCM for a single number of clusters, if input given
    then only improves the input.

******************************************************************/

#define ProgName       "CBFCM"
#define VersionNumber  "Version 0.1.3"
#define LastUpdated    "9.10.2002" /* iak */

#define  FACTFILE  "cbfcm.fac"

#include "parametr.c"
#include "fcm.h"
#include "cb.h"
#include "file.h"
#include "interfc.h"
#include "memctrl.h"
#include "random.h"
#include "cb_util.h"
#include "DistCrit/distance.h"
#include "DistCrit/criteria.h"
#include "sortcb.h"

#define MAXFILENAME 512

#ifndef max
#define max(a,b) ((a) > (b) ? (a) : (b))
#endif

void PrintInfo() {
    PrintMessage("%s\t%s\n(last modified %s, compiled %s, %s)\n\n"
        "Randomized FCM for fixed number of clusters.\n"
        "  Note: produces non-fuzzy partitioning.\n"
        "Usage: %s data.ts out.cb [out.pa [in.cb [in.pa]]] %coptions\n\n",
        ProgName, VersionNumber, LastUpdated, __DATE__, __TIME__,
	ProgName, OPTION_SYMBOL);
    PrintOptions();
    PrintMessage("\n");
}

void PrintOperatingInfo() {
    if (Value(QuietLevel)) {
        PrintMessage("\n%s %s %s\n\n", ProgName, VersionNumber, LastUpdated);
        PrintSelectedOptions();
    }
}

static char DataName[MAXFILENAME] = "\0";
static char OutCBName[MAXFILENAME] = "\0";
static char OutPAName[MAXFILENAME] = "\0";
static char InCBName[MAXFILENAME] = "\0";
static char InPAName[MAXFILENAME] = "\0";

#define PARAMETER_FILENAME_COUNT 5

static ParameterInfo paraminfo[PARAMETER_FILENAME_COUNT] = { 
    { DataName, "", 0, INFILE },
    { OutCBName, "", 0, OUTFILE },
    { OutPAName, "", 1, OUTFILE },
    { InCBName, "", 2, INFILE },
    { InPAName, "", 3, INFILE } };

void CheckParameters() {
    if (Value(Distance) == EvaluationFunctionDefault) SetValue(Distance,
	ciCriterionDefaultDistance(Value(EvaluationFunction)));
    if (ExistFile(OutCBName) && !Value(Overwrite)) {
	ErrorMessage("Output codebook exists and no overwrite specified.\n");
	ExitProcessing(3);
    }
    if (ExistFile(OutPAName) && !Value(Overwrite)) {
	ErrorMessage("Output partition exists and no overwrite specified.\n");
	ExitProcessing(4);
    }
    if (*InCBName && !ExistFile(InCBName)) {
	ErrorMessage("Input codebook not found.\n");
	ExitProcessing(5);
    }
    if (*InPAName && !ExistFile(InPAName)) {
	ErrorMessage("Input partition not found.\n");
	ExitProcessing(6);
    }
    if (!*DataName || !ExistFile(DataName)) {
	ErrorMessage("Input data file not found.\n");
	ExitProcessing(7);
    }
}

static void Finalize(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA);

static float MinimumValue = 0;
static float MaximumValue = 0;
static int FileFormat = 0;

static void Initialize(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA) {
    float mmin, mmax;
    int foo;
    if (ReadGenericData(DataName, TS,
	&FileFormat, &MinimumValue, &MaximumValue))
    {
	ErrorMessage("Failed to read data.\n");
	ExitProcessing(12);
    }
    if (*InCBName) {
	if (ReadGenericData(InCBName, CB, &foo, &mmin, &mmax)) {
	    ErrorMessage("Failed to read centroids.\n");
	    ExitProcessing(13);
	}
    } else { /* do random solution */
	/* if no input codebook given, no partitioning either */
	if (FileFormat != RGD_CB && !*OutPAName) {
	    if (Value(QuietLevel) > 1) PrintMessage(
		"Data vectors before duplicate removal: %i,", BookSize(TS));
	    RemoveDuplicatesFromCodebook(TS);
	    if (Value(QuietLevel) > 1)
		PrintMessage(" after: %i.\n", BookSize(TS));
	}
	CreateNewCodebook(CB, Value(NumberOfClusters), TS);
    }
    if (BookSize(TS) < BookSize(CB)) {
        if (Value(ClampBookSize)) {
            CreateNewCodebook(CB, BookSize(TS), TS);
            CopyCodebook(TS, CB);
            CreateNewPartitioning(PA, TS, BookSize(CB));
            if (*OutPAName) {
                PutAllInOwnPartition(TS, PA);
                GenerateOptimalPartitioning(TS, CB, PA);
            }
            Finalize(TS, CB, PA);
            ExitProcessing(0);
        } else {
	    ErrorMessage("More clusters than data vectors.\n");
	    ExitProcessing(10);
        }
    }
    if (*InPAName) {
	if (ReadGenericMapping(InPAName, TS, PA, &foo)) {
	    ErrorMessage("Failed to read mapping.\n");
	    ExitProcessing(14);
	}
    } else CreateNewPartitioning(PA, TS, BookSize(CB));
    if (PartitionCount(PA) != BookSize(CB)) {
	ErrorMessage("Partition and codebook sizes do not match.\n");
	ExitProcessing(11);
    }
}

static void Finalize(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA) {
    if (!*OutPAName) SortCodebook(CB, DATA_ASCENDING);
    else WriteGenericMapping(OutPAName, TS, PA, FileFormat, Value(Overwrite));
    WriteGenericData(OutCBName, CB,
        FileFormat, MinimumValue, MaximumValue, Value(Overwrite));
    FreeCodebook(TS);
    FreeCodebook(CB);
    FreePartitioning(PA);
}

int main(int argc, char** argv) {
    TRAININGSET TS;
    CODEBOOK CB;
    PARTITIONING PA;
    DistanceInfo* DI;
    CriterionInfo* CI;
    float Error, AverageIterations;
    int Sum, GLAIter = -1;

    ParseParameters(argc, argv, PARAMETER_FILENAME_COUNT, paraminfo);
    initrandom(Value(RandomNumberSeed));
    CheckParameters();
    PrintOperatingInfo();
    Initialize(&TS, &CB, &PA);
    DI = diNew(&TS, Value(Distance), 0, 0, 0, 0);
    CI = ciNew(&TS, DI, Value(EvaluationFunction), 0, 0, 0, 0, 0, 0);

    if (Value(FinishWithGLA)) GLAIter = Value(GLAIterations);

    if (*InCBName) {
        Sum = CrispFuzzyCMeans(&TS, &CB, &PA, Value(FCMIterations),
            CI, GLAIter);
        Error = ciEvaluate(CI, &CB, &PA);
    } else {
        Sum = CrispRandomizedFuzzyCMeans(&TS, &CB, &PA, CI,
            &Error, Value(InitialSolutions), Value(FCMIterations),
            GLAIter, max(0, Value(QuietLevel) - 2),
            Value(TraceData), Value(TraceDataID));
    }

    AverageIterations = Sum;
    if (!*InCBName) AverageIterations /= Value(InitialSolutions);
    if (Value(QuietLevel)) {
        if (*InCBName || (Value(QuietLevel) < 3))
            PrintMessage("\nError = %.6g, average iterations = %.3f.\n\n",
                Error, AverageIterations);
        else PrintMessage(
            "\nAverage iterations = %.3f.\n\n", AverageIterations);
    }
    ciDelete(CI);
    diDelete(DI);
    Finalize(&TS, &CB, &PA);
    return 0;
}

