#include "fuzzy.hpp"

void FuzzyPartitioning::Fuzzify(PARTITIONING* PA) {
    M.Fill(0.0f);
    for (int n = 0; n < Vectors(); ++n) M(n, Map(PA, n)) = 1.0f;
}

FuzzyPartitioning::FuzzyPartitioning(TRAININGSET* TS, int Groups,
    float FuzzyPower) : power(FuzzyPower), M(BookSize(TS), Groups), dataset(TS)
{
    M.Fill(0.0f);
}

FuzzyPartitioning::FuzzyPartitioning(TRAININGSET* TS, CODEBOOK* CB,
    PARTITIONING* PA, float FuzzyPower)
        : power(FuzzyPower), M(BookSize(TS), BookSize(CB)), dataset(TS)
{
    Fuzzify(PA);
}

const float MinimumDistance = 1e-6;

/* Dunn's paper Step 3 in section 6. */
void FuzzyPartitioning::UpdateMemberships(CODEBOOK* CB, DistanceInfo* DI) {
    for (int n = 0; n < Vectors(); ++n) {
        int MinIndex = 0;
        float MinDist = 1e30;
        double InvDistSum = 0;
        RowIterator End = M.RowEnd(n);
        RowIterator Iter = M.RowBegin(n);
        for (int k = 0; k < Groups(); ++k) {
            float Dist = diDistance(DI, dataset, CB, 0, n, k, DT_DATA2MODEL);
            if (DI->Type != DT_EuclideanSquare) Dist *= Dist;
            if (Dist < MinDist) {
                MinIndex = k;
                MinDist = Dist;
            }
            if (Dist > MinimumDistance) {
                *Iter = 1.0f / Dist;
                //M(n, k) = 1.0f / Dist;
                InvDistSum += *Iter; //M(n, k);
                ++Iter;
            }
        }
        if (MinDist < MinimumDistance) { // Point too close.
            for (Iter = M.RowBegin(n); Iter != End; ++Iter) *Iter = 0.0f;
            //for (int k = 0; k < Groups(); ++k) M(n, k) = 0;
            M(n, MinIndex) = 1;
        } else
            for (Iter = M.RowBegin(n); Iter != End; ++Iter) *Iter /= InvDistSum;
            //for (int k = 0; k < Groups(); ++k) M(n, k) /= InvDistSum;
    }
}

void FuzzyPartitioning::Crispify(PARTITIONING* PA) {
    for (int n = 0; n < Vectors(); ++n) {
        int Closest = 0;
        for (int k = 1; k < Groups(); ++k)
            if ((*this)(n, k) > (*this)(n, Closest)) Closest = k;
        if (Map(PA, n) != Closest) ChangePartition(dataset, PA, Closest, n);
    }
}


FuzzyPartitioning* fpNew(TRAININGSET* TS, int Groups, float FuzzyPower) {
    FuzzyPartitioning* FPA = new FuzzyPartitioning(TS, Groups, FuzzyPower);
    return FPA;
}

void fpDelete(FuzzyPartitioning* FPA) {
    if (!FPA) return;
    delete FPA;
}

void fpUpdateMemberships(FuzzyPartitioning* FPA, CODEBOOK* CB,
    DistanceInfo* DI)
{
    FPA->UpdateMemberships(CB, DI);
}

float fpMembership(FuzzyPartitioning* FPA, int Vector, int Group) {
    return (*FPA)(Vector, Group);
}

int fpVectors(FuzzyPartitioning* FPA) {
    return FPA->Vectors();
}

int fpGroups(FuzzyPartitioning* FPA) {
    return FPA->Groups();
}


FuzzyPartitioning* Fuzzify(TRAININGSET* TS, PARTITIONING* PA,
    FuzzyPartitioning* FPA)
{
    if (!FPA) FPA = new FuzzyPartitioning(TS, PartitionCount(PA), 2.0f);
    FPA->Fuzzify(PA);
    return FPA;
}


void Crispify(FuzzyPartitioning* FPA, PARTITIONING* PA) {
    FPA->Crispify(PA);
}


